/**
 * @modification $Date: 2001/12/03 16:28:07 $
 * @statut       unstable
 * @file         BuCheckBox3States.java
 * @version      0.36
 * @author       Bertrand Marchand
 * @email        guillaume@desnoix.com
 * @license      GNU General Public License 2 (GPL2)
 * @copyright    1998-2001 Guillaume Desnoix
 * (c)1998-2001 Bertrand Marchand
 */

package com.memoire.bu;

import com.memoire.bu.*;
import com.memoire.dnd.*;
import com.memoire.fu.*;
import com.memoire.re.*;


/**
 * @creation     15/12/98
 * @modification 05/01/99
 * @statut       instable
 */

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

/**
 * Un checkbox permettant 3 tats (activ, dsactiv, mixte). Dans ce dernier
 * cas, le bouton apparat en gris
 * @version 0.02, 05/01/99
 * @author Bertrand Marchand
 */
public class BuCheckBox3States extends JCheckBox {
  public static final int STATE_MIXED      = 0;                    // On reprend les status Item pour rester compatible avec la mthode
  public static final int STATE_SELECTED   = ItemEvent.SELECTED;   // ItemEvent.getStateChanged(). On rajoute un tat BuCheckBox3States.STATE_MIXED.
  public static final int STATE_DESELECTED = ItemEvent.DESELECTED;

  private int state_;                      // Etat (SELECTED, DESELECTED, MIXED)
  private Icon iconStateMixed = null;      // L'icon de status mixte (pourrait par la suite tre issu de iconDefault).
  private Icon iconDefault = null;         // L'icon donn par le programmeur
  private Vector itemListenerList = null;  // Les items listener du composant
  private boolean defaultEvent_;           // Definit si l'venement ItemEvent provient d'une des classes mres


  /**
   * Creates an initially deselected checkbox button with no text, no icon.
   */
  public BuCheckBox3States () {
    this(null, null, STATE_DESELECTED);
  }

  /**
   * Creates an initially deselected checkbox with an icon.
   *
   * @param icon  the Icon image to display
   */
  public BuCheckBox3States(Icon icon) {
    this(null, icon, STATE_DESELECTED);
  }

  /**
   * Creates a checkbox with an icon and specifies its state.
   *
   * @param icon  the Icon image to display
   * @param state an int value indicating the initial selection
   *        state. The possible values are STATE_DESELECTED,
   *        STATE_SELECTED, STATE_MIXED.
   */
  public BuCheckBox3States(Icon icon, int state) {
    this(null, icon, state);
  }

  /**
   * Creates an initially deselected checkbox with text.
   *
   * @param text the text of the checkbox.
   */
  public BuCheckBox3States (String text) {
    this(text, null, STATE_DESELECTED);
  }

  /**
   * Creates a checkbox with text and specifies its state.
   *
   * @param text the text of the checkbox.
   * @param state an int value indicating the initial selection
   *        state. The possible values are STATE_DESELECTED,
   *        STATE_SELECTED, STATE_MIXED.
   */
  public BuCheckBox3States (String text, int state) {
    this(text, null, state);
  }

  /**
   * Creates an initially deselected checkbox with
   * the specified text and icon.
   *
   * @param text the text of the checkbox.
   * @param icon  the Icon image to display
   */
  public BuCheckBox3States(String text, Icon icon) {
    this(text, icon, STATE_DESELECTED);
  }

  /**
   * Creates a checkbox with the specified text and icon and specifies its state.
   *
   * @param text the text of the checkbox.
   * @param icon  the Icon image to display
   * @param state an int value indicating the initial selection
   *        state. The possible values are STATE_DESELECTED,
   *        STATE_SELECTED, STATE_MIXED.
   */
  public BuCheckBox3States(String text, Icon icon, int state) {
    super(text, icon);
    iconDefault = icon;
    iconStateMixed = new IconStateMixed();
    itemListenerList = new Vector();
    defaultEvent_ = true;
    this.setState(state);
  }

  /**
   * Affectation du status (SELECTED, DESELECTED, MIXED)
   */
  public void setState(int state) {
    defaultEvent_ = false;
    int vp = state_;
    state_ = state;
    switch (state) {
     case STATE_SELECTED:
      setIcon(iconDefault);
      setSelected(true);
      break;
     case STATE_DESELECTED:
      setIcon(iconDefault);
      setSelected(false);
      break;
     case STATE_MIXED:
      setIcon(iconStateMixed);
      fireItemStateChanged(new ItemEvent(this,ItemEvent.ITEM_STATE_CHANGED,this,
                           state));
      fireStateChanged();
      break;
    }
    defaultEvent_ = true;
  }

  /**
   * Retourne le status (SELECTED, DESELECTED, MIXED)
   */
  public int getState() {
    return state_;
  }

  /**
   * adds an ItemListener to the checkbox
   */
  public void addItemListener(ItemListener l) {
    itemListenerList.addElement(l);
  }

  /**
   * removes an ItemListener from the button
   */
  public void removeItemListener(ItemListener l) {
    itemListenerList.removeElement(l);
  }

  protected void fireItemStateChanged(ItemEvent event) {
    // Evenement en provenance des classes mres => On ne notifie pas les listeners, mais
    // on remodifie l'tat du bouton.
    if (defaultEvent_) {
      setState((state_+1)%3);
      return;
    }
    ItemEvent evt = new ItemEvent(this,event.getID(),event.getItem(),event.getStateChange());
    if (itemListenerList != null)
    for (int i=0; i<itemListenerList.size(); i++) {
      ((ItemListener) itemListenerList.elementAt(i)).itemStateChanged(evt);
    }
  }
  /**
   * Doit tre remplac par <I>setState</I>
   */
//  public void setSelected(boolean state) {
//    throw new RuntimeException("BuCheckBox3States : utilisez 'setState' a la place de 'setSelected'");
//  }

  /**
   * Doit tre remplac par <I>getState</I>
   */
//  public boolean isSelected() {
//    throw new RuntimeException("BuCheckBox3States : utilisez 'getState' a la place de 'isSelected'");
//  }
}

// Classe de trace de l'icon pour un status STATE_MIXED

class IconStateMixed implements Icon {

  public int getIconHeight() {
    return 13;
  }

  public int getIconWidth() {
    return 13;
  }

  public void paintIcon(Component _c, Graphics _g, int _x, int _y) {
    int w=getIconWidth();
    int h=getIconHeight();

    _g.setColor(Color.lightGray);
    _g.fillRect(_x,_y,w-2,h-1);
    _g.setColor(Color.gray);
    _g.drawLine(_x+w-1,_y,_x,_y);
    _g.drawLine(_x,_y,_x,_y+h-1);
    _g.setColor(Color.white);
    _g.drawLine(_x,_y+h-1,_x+w-1,_y+h-1);
    _g.drawLine(_x+w-1,_y+h-1,_x+w-1,_y);

    _g.setColor(Color.gray);
    _g.drawLine(_x+3,_y+3,_x+w-5,_y+h-5);
    _g.drawLine(_x+4,_y+3,_x+w-5,_y+h-6);
    _g.drawLine(_x+3,_y+4,_x+w-6,_y+h-5);
    _g.drawLine(_x+3,_y+h-5,_x+w-5,_y+3);
    _g.drawLine(_x+3,_y+h-6,_x+w-6,_y+3);
    _g.drawLine(_x+4,_y+h-5,_x+w-5,_y+4);
  }
}
