/**
 * @product      Alma (Atelier logiciel de modelistion et d'analyse)
 * @file         ScriptInterpreterRuby.java
 * @version      0.36
 * @www          http://www.memoire.com/guillaume-desnoix/alma/
 * @modification $Date: 2001/12/03 16:28:08 $
 * @author       Guillaume Desnoix
 * @email        guillaume@desnoix.com
 * @license      GNU General Public License 2 (GPL2)
 */

// Contrib: Jan Arne Pertersen
package com.memoire.script;

import java.io.*;
import java.lang.reflect.*;

import org.jruby.*;
import org.jruby.javasupport.*;
import org.jruby.regexp.*;

public class ScriptInterpreterRuby implements ScriptInterpreterInterface {
    private Ruby ruby;

    public ScriptInterpreterRuby() {
        try {
            Class.forName("org.jruby.Ruby");
        } catch (ClassNotFoundException ex) {
            throw new RuntimeException("Ruby not found");
        }
    }

    public String getName() {
        return "Ruby";
    }

    public String getShortDescription() {
        return "A ruby interpreter";
    }

    public String getLongDescription() {
        return "Ruby is a ruby interpreter.";
    }

    public Object getIcon() {
        return null;
    }

    public String getExtension() {
        return "rb";
    }

    public String[] getExtensions() {
        return new String[] { "rb", "RB" };
    }

    public Object getGUI() {
        return null;
    }

    public String toString() {
        return "interpreter " + getName();
    }

    private InputStream in_;
    private OutputStream out_, err_;

    private LineNumberReader input;
    private PrintWriter output, error;

    public void setInputStream(InputStream _in) {
        in_ = _in;
        input = new LineNumberReader(new InputStreamReader(_in));
    }

    public InputStream getInputStream() {
        return in_;
    }

    public void setOutputStream(OutputStream _out) {
        out_ = _out;
        output = new PrintWriter(_out, true);
    }

    public OutputStream getOutputStream() {
        return out_;
    }

    public void setErrorStream(OutputStream _err) {
        err_ = _err;
        error = new PrintWriter(_err, true);
    }

    public OutputStream getErrorStream() {
        return err_;
    }

    public void start() {
        if (in_ == null)
            setInputStream(System.in);
        if (out_ == null)
            setOutputStream(System.out);
        if (err_ == null)
            setErrorStream(System.err);

        ruby = new Ruby();
        // JAP: You have now to set the Regexp Adapter class (GNU or Java 1.4).
        ruby.setRegexpAdapterClass(GNURegexpAdapter.class);
        ruby.init();

        // JAP: Added 20.11.2001 to JRuby.
        ruby.getRuntime().setInputStream(getInputStream());
        ruby.getRuntime().setOutputStream(new PrintStream(getOutputStream(),true));
        ruby.getRuntime().setErrorStream(new PrintStream(getErrorStream(),true));
    }

    public void run() {
        boolean hitEOF = false;
        while (!hitEOF) {
            int startline = input.getLineNumber();
            error.flush();
            output.print("ruby> ");
            output.flush();
            try {
                String source = "";
                while (true) {

                    while (!input.ready()) {
                        try {
                            Thread.sleep(100);
                        } catch (InterruptedException ex) {
                        }
                        if (exited_) {
                            hitEOF = true;
                            break;
                        }
                    }

                    String newline = input.readLine();

                    if (newline == null) {
                        hitEOF = true;
                        break;
                    }

                    newline = newline.trim();

                    if (!newline.endsWith("\\")) {
                        source = source + newline;
                        break;
                    } else {
                        newline = newline.substring(0, newline.length() - 1);
                        source = source + newline + " ";
                        output.print("      ");
                        output.flush();
                    }
                }

                if (!"".equals(source.trim())) {
                    RubyString rs = RubyString.m_newString(ruby, source);

                    RubyObject result =
                        ruby.getRubyTopSelf().eval(
                            ruby.getRubyParser().compileString("<stdin>", rs, 0));

                    if ((result != null) && (result != ruby.getNil()))
                        output.println(
                            "      " + JavaUtil.convertRubyToJava(ruby, result, Object.class));
                }
            } catch (Exception ex) {
                error.println(ex.getClass().getName()+": "+ex.getMessage());
            }
        }

	output.flush();
	error.flush();
    }

    private boolean exited_ = false;

    public void exit() {
        exited_ = true;
    }

    public void eval(String _script) {
        if ("".equals(_script.trim()))
            return;

        try {
            RubyString rs = RubyString.m_newString(ruby, _script);

            RubyObject result =
                ruby.getRubyTopSelf().eval(
                    ruby.getRubyParser().compileString("<script>", rs, 0));

            if (result != ruby.getNil())
                output.println(JavaUtil.convertRubyToJava(ruby, result, Object.class));
        } catch (Exception ex) {
            error.println(ex.getMessage());
        }

	output.flush();
	error.flush();
    }

    public void addVariable(String _nom, Object _objet) {
        // JAP: Global variables have to start with "$" in Ruby
        if (!_nom.startsWith("$")) {
            _nom = "$" + _nom;
        }
        ruby.setGlobalVar(
            _nom,
            JavaUtil.convertJavaToRuby(
                ruby,
                _objet,
                (_objet != null ? _objet.getClass() : null)));
    }

    public void removeVariable(String _nom) {
        // JAP: Global variables have to start with "$" in Ruby
        if (!_nom.startsWith("$")) {
            _nom = "$" + _nom;
        }

        ruby.setGlobalVar(_nom, ruby.getNil());
    }

    public void addFunction(String _name, Method _method) {
        // JAP: I use the same class which is used by the Javasupport of JRuby
        RubyClass objectClass = ruby.getClasses().getObjectClass();
        objectClass.defineMethod(_name, new JavaMethod(new Method[] { _method }));
    }

    public void removeFunction(String _name) {
        RubyClass objectClass = ruby.getClasses().getObjectClass();
        objectClass.undefMethod(_name);
    }

    //////////////// Main Loop

    public static void main(String[] files) {
        ScriptInterpreterRuby ruby = new ScriptInterpreterRuby();
        ruby.start();

        /*
        try
        {
          for(int i=0; i<files.length;i++) ruby.load(files[i]);
        }
        catch (RuntimeException ex) { }
        */
        ruby.run();
    }
}
