

package com.memoire.silk;
import com.memoire.silk.*;


import java.io.*;

/** This class represents a SilkScheme interpreter.
 * See http://www.norvig.com/SILK.html for more documentation.
 * This is version 1.4. 
 * @author Peter Norvig, peter@norvig.com http://www.norvig.com 
 * Copyright 1998 Peter Norvig, see http://www.norvig.com/license.html
 * Adapted by Guillaume Desnoix for ScriptInterpreterInterface.
 **/

public class SilkScheme
       extends SilkSchemeUtils
{
  public SilkInputPort   input;
  public PrintWriter     output,error;
  public SilkEnvironment globalSilkEnvironment;

  /** Create a SilkScheme interpreter and load an array of files into it.
   * Also load SilkSchemePrimitives.CODE. **/

  public SilkScheme()
  {
  }

  /** @GDX */
  public void exit() { input.exited_=true; }

  /** Prompt, read, eval, and write the result. 
   * Also sets up a catch for any RuntimeExceptions encountered. **/
  public void readEvalWriteLoop() {
    Object x;
    for(;;) {
      try {
	output.print("> "); output.flush();
	if (input.isEOF(x = input.read())) return;
	write(eval(x), output, true); 
	output.println(); output.flush();
      } catch (RuntimeException e) { }
    }
  }

  /** Eval all the expressions in a file. Calls load(SilkInputPort). **/
  public Object load(Object fileName) {
    String name = stringify(fileName, false);
    try { return load(new SilkInputPort(new FileInputStream(name))); }
    catch (IOException e) { return error("can't load " + name); }
  }

  /** Eval all the expressions coming from an SilkInputPort. **/
  public Object load(SilkInputPort in) {
    Object x = null;
    for(;;) {
      if (in.isEOF(x = in.read())) return TRUE;
      eval(x); 
    }
  }
      
  //////////////// Evaluation

  /** Evaluate an object, x, in an environment. **/
  public Object eval(Object x, SilkEnvironment env) { 
    // The purpose of the while loop is to allow tail recursion.
    // The idea is that in a tail recursive position, we do "x = ..."
    // and loop, rather than doing "return eval(...)".
    while (true) {
      if (x instanceof String) {         // VARIABLE
	return env.lookup((String)x);
      } else if (!(x instanceof SilkPair)) { // CONSTANT
	return x;
      } else {                           
	Object fn = first(x);
	Object args = rest(x);
	if (fn == "quote") {             // QUOTE
	  return first(args);
	} else if (fn == "begin") {      // BEGIN
	  for (; rest(args) != null; args = rest(args)) {
	    eval(first(args), env);
	  }
	  x = first(args);
	} else if (fn == "define") {     // DEFINE
	  if (first(args) instanceof SilkPair)
	    return env.define(first(first(args)),
	     eval(cons("lambda", cons(rest(first(args)), rest(args))), env));
	  else return env.define(first(args), eval(second(args), env));
	} else if (fn == "set!") {       // SET!
	  return env.set(first(args), eval(second(args), env));
	} else if (fn == "if") {         // IF
	  x = (truth(eval(first(args), env))) ? second(args) : third(args);
	} else if (fn == "cond") {       // COND
	  x = reduceCond(args, env);
	} else if (fn == "lambda") {     // LAMBDA
	  return new SilkClosure(first(args), rest(args), env);
	} else if (fn == "macro") {      // MACRO
	  return new SilkMacro(first(args), rest(args), env);
	} else {                         // PROCEDURE CALL:
	  fn = eval(fn, env);
	  if (fn instanceof SilkMacro) {          // (MACRO CALL)
	    x = ((SilkMacro)fn).expand(this, (SilkPair)x, args);
	  } else if (fn instanceof SilkClosure) { // (CLOSURE CALL)
	    SilkClosure f = (SilkClosure)fn;
	    x = f.body;
	    env = new SilkEnvironment(f.parms, evalList(args, env), f.env);
	  } else {                            // (OTHER PROCEDURE CALL)
	    return SilkProcedure.proc(fn).apply(this, evalList(args, env));
	  }
	}
      }
    }
  }

  /** Eval in the global environment. **/
  public Object eval(Object x) { return eval(x, this.globalSilkEnvironment); }

  /** Evaluate each of a list of expressions. **/
  SilkPair evalList(Object list, SilkEnvironment env) {
    if (list == null) 
      return null;
    else if (!(list instanceof SilkPair)) {
      error("Illegal arg list: " + list);
      return null;
    } else 
      return cons(eval(first(list), env), evalList(rest(list), env));
  }

  /** Reduce a cond expression to some code which, when evaluated,
   * gives the value of the cond expression.  We do it that way to
   * maintain tail recursion. **/
  Object reduceCond(Object clauses, SilkEnvironment env) {
    Object result = null;
    for (;;) {
      if (clauses == null) return FALSE;
      Object clause = first(clauses); clauses = rest(clauses);
      if (first(clause) == "else" 
	  || truth(result = eval(first(clause), env)))
	if (rest(clause) == null) return list("quote", result);
	else if (second(clause) == "=>")
	  return list(third(clause), list("quote", result));
	else return cons("begin", rest(clause));
    }
  }
}


